// PrOF 2018-2020
// Thanks to Derpduck as this HUD is updated from an old version of his Esport mod
// Major thanks to YB for helping with the 2018 version. Some networking learned from Lorgan. Completely remade in 2020

/*
Mini dev blog:
- I miss short-circuit evaluation
- And the ternary operator
- Tbh I also miss scripts but user events are pretty close
*/
/*
- Modes:
   - HOST_AUTHORITY, only host can start
   - MIN_PLAYERS, start if everyone is ready AND there are at least X non spectators
   - MIN_TIME, start if everyone is ready AND X time passed
*/

global.rup_packetID = argument1;
global.rup_buffer = buffer_create();

global.rup_matchStarted = false;
global.playersWeCareAbout = ds_list_create();

// This gets executed as soon as the plugin is dowloaded, so just before match start for the host
global.rup_mode = 0;
global.rup_countSpectators = false;
global.rup_minPlayers = 0;
global.rup_minTime = 0;
if(global.isHost) {

	global.rup_mode = show_message_ext("-----ReadyUp-----#Choose mode:", "HOST_AUTHORITY", "MIN_PLAYERS", "MIN_TIME");
	if (global.rup_mode == 0) global.rup_mode = 1;

	if (global.rup_mode == 2) {
		global.rup_minPlayers = get_integer("-----ReadyUp-----#Minimum players before allowing game to start:", 2);
	}
	else if (global.rup_mode == 3) {
		global.rup_minTime = get_integer("-----ReadyUp-----#Minimum time in seconds before allowing game to start:", 30);
	}

	global.rup_countSpectators = show_question("-----ReadyUp-----#Allow spectators to ready up?");	
}

// DEBUG
/*
object_event_add(Character, ev_draw, 0, '
	draw_set_color(c_black);
	draw_set_halign(fa_center);
	draw_set_valign(fa_top);
	draw_text(x, y - 20, "Ready: [" + string(player.isRupReady) + "]");
	draw_text(x, y + 20, "Started: [" + string(global.rup_matchStarted) + "]");
	draw_set_color(c_white);
	draw_set_halign(fa_left);
');
*/

global.playerInit = '
	isRupReady = false;
';
// The first Player object is created before plugins are loaded
with (Player) execute_string(global.playerInit);

// Remove player from list on disconnect
object_event_add(Player, ev_create, 0, global.playerInit);
object_event_add(Player, ev_destroy, 0, '
	ds_list_delete(global.playersWeCareAbout, ds_list_find_index(global.playersWeCareAbout, id));
');

global.ReadyUpHUD = object_add();
// Ensure we draw over everything else but not over menus
object_set_depth(global.ReadyUpHUD, -110000);
object_set_persistent(global.ReadyUpHUD, false);

// Create
object_event_add(global.ReadyUpHUD, ev_create, 0, '

	// Constants
	RUP_TOGGLEREADY = 0;
	RUP_READY =       1;
	RUP_NOT_READY =   2;
	RUP_START =       3;
	RUP_JOIN =        4;
	RUP_SYNC =        5;

	// Modes
	HOST_AUTHORITY = 1;
	MIN_PLAYERS =    2;
	MIN_TIME =       3;

	// Event descriptions
	EV_RECEIVE_HOST =   0;
	EV_TOGGLE_READY =   1;
	EV_START =          2;
	EV_RECEIVE_CLIENT = 3;
	EV_REQUEST_JOIN =   4;
	EV_SEND_JOIN =      5;
	EV_REQUEST_SYNC =   6;
	EV_SEND_SYNC =      7;
	EV_RESET_GAME =    15;
	
	aboutToStart = false;	
	
	timerHelper = -1;

	//TODO change this and figure out what it actually does
	rectLinesNumber = 1;

	// Support Map Voting by Machidro
	// Note: this only get executed by the host as intended
	//TODO one of these ifs can probably be removed
	if (variable_global_exists("mapVotingSettings")) {
		with (ControlPoint) {
			if (variable_local_exists("onMapVoting")) {
				// If on a map voting map, disable hud and start immediately
				global.rup_matchStarted = onMapVoting;
			}
		}
	}

	event_user(EV_REQUEST_JOIN);
');

// Reset player status and clear list DONE
object_event_add(global.ReadyUpHUD, ev_destroy, 0, '
	//TODO Half of these are unnecessary
	aboutToStart = false;
	ds_list_clear(global.playersWeCareAbout);
	alarm[0] = -1;
	with (Player) isRupReady = false;
	global.rup_matchStarted = false;
');

// Update playersWeCareAbout, run host logic DONE (NEEDS MIN_TIME)
//TODO MIN_PLAYERS is broken, probably
object_event_add(global.ReadyUpHUD, ev_step, ev_step_begin, '
	
	// Update playersWeCareAbout list
	for(i = 0; i < ds_list_size(global.players); i += 1) {
		player = ds_list_find_value(global.players, i);
		
		// If we do not already have the player AND (they are not a spectator OR we let spectators vote)
		if (!ds_list_find_value(global.playersWeCareAbout,ds_list_find_index(global.players,player)) && (player.team != TEAM_SPECTATOR || global.rup_countSpectators)) {
			ds_list_add(global.playersWeCareAbout, player);
		}
		
		// A player changed team to spectator, remove from list if needed
		if (player.team == TEAM_SPECTATOR && !global.rup_countSpectators) {
			ds_list_delete(global.playersWeCareAbout, ds_list_find_index(global.playersWeCareAbout, player));
		}
	}
	
	// Actual logic, should only be done on the host side
	if (!global.isHost && !aboutToStart) exit;

	var shouldStart;
	shouldStart = true;
	
	if (global.rup_mode == HOST_AUTHORITY) {
		shouldStart = false;
	}
	else if (global.rup_mode == MIN_PLAYERS && ds_list_size(global.playersWeCareAbout) < global.rup_minPlayers) {
		shouldStart = false;
	}
	else if (global.rup_mode == MIN_TIME) {
		//TODO I imagine very few people will want this
		shouldStart = false;
	}

	if (shouldStart) event_user(EV_START);
');

// Stop match from beginning if not ready DONE
object_event_add(global.ReadyUpHUD, ev_step, ev_step_normal, '

	if (global.rup_matchStarted) exit;
	
	event_user(EV_RESET_GAME);
');

// Check for user input DONE
object_event_add(global.ReadyUpHUD, ev_step, ev_step_end, '
	// Reminder: do not put an exit statement anywhere in this event

	if (!global.rup_matchStarted && !aboutToStart && !instance_exists(MenuController)) {
		if (keyboard_check_pressed(global.drop)) {
			event_user(EV_TOGGLE_READY);
		}
		else if (keyboard_check_pressed(vk_f4)) {
			event_user(EV_START);
		}
	}
	
	event_user(EV_RECEIVE_HOST);
	event_user(EV_RECEIVE_CLIENT);
');

// Draw HUD DONE
object_event_add(global.ReadyUpHUD, ev_draw, 0, '

	if (global.rup_matchStarted) exit;

	var xoffset, yoffset, xsize, ysize, xcenter, ycenter, orangeTitle, rectHeightAddition, readyNameDisplay, unreadyNameDisplay, countReady, countUnready, listSize, drawHelper, alphaHelper;

	xoffset = view_xview[0];
	yoffset = view_yview[0];
	xsize = view_wview[0];
	ysize = view_hview[0];
	xcenter = view_xview[0] + view_wview[0]/2;
	ycenter = view_yview[0] + view_hview[0]/2;
	orangeTitle = make_color_rgb(255,102,000);
	listSize = ds_list_size(global.playersWeCareAbout);
	rectHeightAddition = (1+listSize*11)+44; // 1+number of names, each line takes up 11px; 33 is for the titles
	readyNameDisplay = "";
	unreadyNameDisplay = "";
	countReady = 0;
	countUnready = 0;

	for(i = 0; i < listSize; i += 1) {
		player = ds_list_find_value(global.playersWeCareAbout, i);
		if (player != -1) {
			if (player.isRupReady) {
				readyNameDisplay += string(player.name)+"#";
				countReady += 1;
			}
			else {
				unreadyNameDisplay += string(player.name)+"#";
				countUnready += 1;
			}
		}
	}

	//  0: Playing Player
	//  1: Spectating player
	//  2: Playing Host
	//  3: Spectating host
	//  4: Player just joined
	if (global.myself != -1) {
		if (!global.isHost) {
			if (global.myself.team != TEAM_SPECTATOR || global.rup_countSpectators)
				drawHelper = 0;
			else
				drawHelper = 1;
		}
		else if (global.myself != -1){
			if (global.myself.team != TEAM_SPECTATOR || global.rup_countSpectators)
				drawHelper = 2;
			else
				drawHelper = 3;
		}
	}
	else {
		drawHelper = 4;
	}

	alphaHelper = 1;
	// lerp alpha based on time until match start
	// 150 : 1 = 20 : 0
	if (alarm[0] != -1)
		alphaHelper = (alarm[0] * global.delta_factor - 30) / 120;
	
	draw_set_halign(fa_left);
	draw_set_valign(fa_top);
	
	// Rectangle background and outline
	draw_set_color(c_black);
	draw_set_alpha(alphaHelper*0.6);
	rectX1 = xoffset+95;
	rectY1 = yoffset+245;
	rectX2 = xoffset+95+(20*9); 
	rectY2 = yoffset+245+rectHeightAddition;
	draw_rectangle(rectX1, rectY1, rectX2, rectY2, false);
	draw_set_color(c_white);
	draw_set_alpha(alphaHelper*1);
	draw_rectangle(rectX1, rectY1, rectX2, rectY2, true);

	// READY title
	draw_set_color(orangeTitle);
	draw_text(xoffset+100, yoffset+250, "---Ready("+string(countReady)+"/"+string(listSize)+")#");

	// READY player list
	draw_set_color(c_green);
	draw_text(xoffset+99,yoffset+251,"#"+readyNameDisplay);
	draw_text(xoffset+99,yoffset+249,"#"+readyNameDisplay);
	draw_text(xoffset+101,yoffset+251,"#"+readyNameDisplay);
	draw_text(xoffset+101,yoffset+249,"#"+readyNameDisplay);
	draw_set_color(c_white);
	draw_text(xoffset+100,yoffset+250,"#"+readyNameDisplay);


	textHeightAddition1 = ((1+string_count("#",readyNameDisplay))*11)+11;

	// UNREADY title
	draw_set_color(orangeTitle);
	draw_text(xoffset+100,yoffset+250+textHeightAddition1,"---Not Ready("+string(countUnready)+"/"+string(listSize)+")#");

	// UNREADY player list
	draw_set_color(c_red);
	draw_text(xoffset+99,yoffset+251+textHeightAddition1,"#"+unreadyNameDisplay);
	draw_text(xoffset+99,yoffset+249+textHeightAddition1,"#"+unreadyNameDisplay);
	draw_text(xoffset+101,yoffset+251+textHeightAddition1,"#"+unreadyNameDisplay);
	draw_text(xoffset+101,yoffset+249+textHeightAddition1,"#"+unreadyNameDisplay);
	draw_set_color(c_white);
	draw_text(xoffset+100,yoffset+250+textHeightAddition1,"#"+unreadyNameDisplay);

	if (drawHelper == 1 || drawHelper == 4) exit;
	
	// extra information
	draw_set_halign(fa_center);
	draw_set_valign(fa_top);

	// Instructions background values
	draw_set_color(c_black);
	draw_set_alpha(alphaHelper*0.6);
	infoX1 = xoffset+xsize/2-175;
	infoY1 = yoffset+ysize/2+50;
	infoX2 = xoffset+xsize/2+177;
	
	if (drawHelper == 2)
		infoY2 = yoffset+ysize/2+75+(rectLinesNumber*11);
	else
		infoY2 = yoffset+ysize/2+60+(rectLinesNumber*11);

	// Instructions background and outline
	draw_rectangle(infoX1, infoY1, infoX2, infoY2, false);
	draw_set_color(c_white);
	draw_set_alpha(alphaHelper*1);
	draw_rectangle(infoX1, infoY1, infoX2, infoY2, true);

	// Instructions text 
	var instrYOffset;
	instrYOffset = 0;
	if (drawHelper == 2) instrYOffset = 15;
	
	if (drawHelper == 0 || drawHelper == 2) {
		draw_text_transformed(xcenter, ycenter+55, "Press " + chr(global.drop) + " to toggle ready", 1, 1, 0);
	}
	if (drawHelper == 2 || drawHelper == 4) {
		draw_text_transformed(xcenter, ycenter+instrYOffset+55, "Press the F4 key to force game start", 1, 1, 0);
	}
	
	// Allow health and ammo HUDs to draw text properly
	draw_set_valign(fa_middle);
');

// Receive messages as host and reply/dispatch DONE (NEEDS SYNC)
object_event_add(global.ReadyUpHUD, ev_other, ev_user0, '
	if (!global.isHost) exit;
	
	while(PluginPacketGetBuffer(global.rup_packetID) != -1) {

		rupReceiveBuffer = PluginPacketGetBuffer(global.rup_packetID);
		player = PluginPacketGetPlayer(global.rup_packetID);

		switch(read_ubyte(rupReceiveBuffer)) {

			case RUP_TOGGLEREADY:
				player.isRupReady = !player.isRupReady;
				event_user(EV_SEND_JOIN);
				break;

			case RUP_JOIN:
				event_user(EV_SEND_JOIN);
				break;

			case RUP_SYNC:
				//TODO
				break;
		}

		buffer_clear(rupReceiveBuffer);
		PluginPacketPop(global.rup_packetID);
	}
');

// Send isRupReady status change as host or client DONE
object_event_add(global.ReadyUpHUD, ev_other, ev_user1, '
	
	if (global.myself.team == TEAM_SPECTATOR && !global.rup_countSpectators) exit;
	
	if (global.myself.team != TEAM_SPECTATOR) {
		write_ubyte(global.serverSocket, CHAT_BUBBLE);
		write_ubyte(global.serverSocket, global.myself.isRupReady + 43);
	}
	
	if (global.isHost) {

		global.myself.isRupReady = !global.myself.isRupReady;

		if (global.myself.isRupReady) {
			write_ubyte(global.rup_buffer, RUP_READY);
		}
		else {
			write_ubyte(global.rup_buffer, RUP_NOT_READY);
		}
		write_ubyte(global.rup_buffer, 0);

		with(Player) {
			if (id != global.myself) {
				PluginPacketSendTo(global.rup_packetID, global.rup_buffer, id);
			}
		}
		buffer_clear(global.rup_buffer);
	}
	else {
		write_ubyte(global.rup_buffer, RUP_TOGGLEREADY);
		PluginPacketSend(global.rup_packetID, global.rup_buffer);
		buffer_clear(global.rup_buffer);
	}
');

// Start the game as host DONE
object_event_add(global.ReadyUpHUD, ev_other, ev_user2, '

	with (NoticeO) {
		instance_destroy();
	}
	notice = instance_create(0, 0, NoticeO);
	notice.notice = NOTICE_CUSTOM;
	notice.message = "The match is about to begin";

	aboutToStart = true;
	alarm[0] = 150 / global.delta_factor;

	with (Player) isRupReady = true;

	if (!global.isHost) exit;
	
	write_ubyte(global.rup_buffer, RUP_START);
	with(Player) {
		if (id != global.myself) {
			PluginPacketSendTo(global.rup_packetID, global.rup_buffer, id);
		}
	}
	buffer_clear(global.rup_buffer);
');

// Read messages as client DONE (NEEDS SYNC)
object_event_add(global.ReadyUpHUD, ev_other, ev_user3, '
	if (global.isHost) exit;
	
	while(PluginPacketGetBuffer(global.rup_packetID) != -1) {
		rupReceiveBuffer = PluginPacketGetBuffer(global.rup_packetID);
		command = read_ubyte(rupReceiveBuffer);
		switch(command) {

			case RUP_START:
				event_user(EV_START);
				break;
			
			case RUP_JOIN:
				global.rup_matchStarted = read_ubyte(rupReceiveBuffer);
				global.rup_mode = read_ubyte(rupReceiveBuffer);
				global.rup_countSpectators = read_ubyte(rupReceiveBuffer);
				break;

			case RUP_READY:
				var playerId, player;
				playerId = read_ubyte(rupReceiveBuffer);
				player = ds_list_find_value(global.players, playerId);
				player.isRupReady = true;
				//DEBUG
				//show_message("Got READY for player index " + string(playerId));
				break;

			case RUP_NOT_READY:
				var playerId, player;
				playerId = read_ubyte(rupReceiveBuffer);
				player = ds_list_find_value(global.players, playerId);
				player.isRupReady = false;
				//DEBUG
				//show_message("Got NOT_READY for player index " + string(playerId));
				break;

			default:
				show_message("Unknown command in client receive: " + string(command));
		}

		// Note: the plugin system is limited (probably) by design to one command per packet
		buffer_clear(rupReceiveBuffer);
		PluginPacketPop(global.rup_packetID);
	}
');

// Request sync match status as client DONE
object_event_add(global.ReadyUpHUD, ev_other, ev_user4, '
	
	if (global.isHost) exit;
	
	write_ubyte(global.rup_buffer, RUP_JOIN);
	PluginPacketSend(global.rup_packetID, global.rup_buffer);
	buffer_clear(global.rup_buffer);
');

// Sync everything with everyone as host DONE (NEEDS SPLIT)
object_event_add(global.ReadyUpHUD, ev_other, ev_user5, '

	if (!global.isHost) exit;
 
	//TODO split this into 2 events
	write_ubyte(global.rup_buffer, RUP_JOIN);
	write_ubyte(global.rup_buffer, global.rup_matchStarted);
	write_ubyte(global.rup_buffer, global.rup_mode);
	write_ubyte(global.rup_buffer, global.rup_countSpectators);
	
	PluginPacketSendTo(global.rup_packetID, global.rup_buffer, player.id);
	buffer_clear(global.rup_buffer);
	
	for (i = 0; i <  ds_list_size(global.playersWeCareAbout); i += 1) {

		var getPlayer, getPlayerId;
		getPlayer = ds_list_find_value(global.playersWeCareAbout, i);
		getPlayerIndex = ds_list_find_index(global.players, getPlayer);

		if (getPlayer.isRupReady) {
			write_ubyte(global.rup_buffer, RUP_READY);
		}
		else {
			write_ubyte(global.rup_buffer, RUP_NOT_READY);			
		}
		write_ubyte(global.rup_buffer, getPlayerIndex);

		//DEBUG
		//show_message("Wrote in buffer READY/NOT_READY for player index " + string(getPlayerIndex));

		// Ensure we only send one command per packet
		with (Player) {
			if (id != global.myself) {
				PluginPacketSendTo(global.rup_packetID, global.rup_buffer, id);
			}
		}
		buffer_clear(global.rup_buffer);
	}	
');

// Reset timers and game objects DONE (NEEDS CHECKING)
object_event_add(global.ReadyUpHUD, ev_other, ev_user15, '

	// Refill game timers based on game mode.
	// Prevent Intelligences from being captured
	// Prevent Generators from being destroyed
	//TODO check for compatibility with 2.9.2
	if (instance_exists(CTFHUD)){
		CTFHUD.timer = CTFHUD.timeLimit;
		GameServer.syncTimer = 1;
		global.blueCaps = 0;
		global.redCaps = 0;
	}
	if (instance_exists(ControlPointHUD)){
		if (timerHelper == -1) timerHelper = global.setupTimer;
		global.setupTimer = timerHelper;
		if (ControlPointHUD.mode == 1){
			ControlPointHUD.timer = min(2*30*60,ControlPointHUD.timeLimit);
		}
		else{
			ControlPointHUD.timer = ControlPointHUD.timeLimit;
		}
		GameServer.syncTimer = 1;
	}
	if (instance_exists(KothHUD)){
		KothHUD.cpUnlock = 900;
		GameServer.syncTimer = 1;
	}
	if (instance_exists(DKothHUD)){
		DKothHUD.cpUnlock = 900;
		GameServer.syncTimer = 1;
	}
	if (instance_exists(GeneratorHUD)){
		GeneratorHUD.timer = GeneratorHUD.timeLimit;
		GameServer.syncTimer = 1;
		with (Generator) {
			shieldHp = maxShieldHp;
			hp = maxHp;
		}
	}
	if (instance_exists(ArenaHUD)){
		ArenaHUD.timer = ArenaHUD.timeLimit;
		ArenaHUD.cpUnlock = 45*30;
		GameServer.syncTimer = 1;
		//serverArenaRestart();
		//TODO Check if this should be done
	}
	if (instance_exists(TeamDeathmatchHUD)){
		TeamDeathmatchHUD.timer = TeamDeathmatchHUD.timeLimit;
		global.redCaps = 0;
		global.blueCaps = 0;
		GameServer.syncTimer = 1;
	}
	if (instance_exists(InvasionHUD)){
		InvasionHUD.timer = InvasionHUD.timeLimit;
		if (timerHelper == -1) timerHelper = global.setupTimer;
		global.setupTimer = timerHelper;
		GameServer.syncTimer = 1;
	}
');

// Countdown to start for host DONE (NEEDS CHECKING)
object_event_add(global.ReadyUpHUD, ev_alarm, 0, '

	// This is the only thing that is done for both client and host
	global.rup_matchStarted = true;

	if (!global.isHost) exit;

	// Reset character stuff
	with (Character) {
		x = xstart;
		y = ystart;
		hp = maxHp;
		ubered = 0;
		hspeed = 0;
		vspeed = 0;
		speed = 0;
		moveStatus = 0;
	}
	with (Medigun) {
		ubering = false;
		uberCharge = 0;
		uberReady = false;
	}
	with (Weapon) {
		// Refill ammo on game start
		ammoCount = maxAmmo;
	}

	// Reset timers and objectives one last time
	event_user(EV_RESET_GAME);
');

// GM decided it will not run the ReadyUpHud create event unless I do this horribleness
// On the upside, this way I cans recreate the object on each map start
object_event_add(RoomChangeObserver, ev_other, ev_room_start, '
	if (room == CustomMapRoom && !instance_exists(global.ReadyUpHUD)) {
		instance_create(0, 0, global.ReadyUpHUD);
	}
');
